c	***********************************************************
c	function to get beta from the initialized 2-D field
c	( deriv is the derivative with respect to a )
c	***********************************************************
	real function getbeta(a,b,deriv)
	implicit none
	integer ma,mb

	include 'fsizes.inc'
	real field(mx,mx),a0,da,b0,db,a,b,deriv
	real la,findxa,findxb,fwgta,fwgtb,ln10,point1,point2
	integer indxa,indxb

	common /betafeld/ a0,da,ma,b0,db,mb,field

c	ln(10) for derivatives
	data ln10 /2.302585/

c	*************************************************************
c	Bilinear interpolation
c	*************************************************************
c	grid place in a
	la=alog10(a)
        findxa=(la-a0)/da+1.0
        indxa=findxa
c	Allow extrapolation for a only
        if (indxa.lt.1) then
          indxa=1
	else if (indxa.ge.ma) then
	  indxa=ma-1
        endif
        fwgta=findxa-indxa
c	grid place in b
        findxb=(b-b0)/db+1.0
        indxb=findxb
        if (indxb.lt.1) then
          indxb=1
          fwgtb=0.0
        else if (indxb.ge.mb) then
          indxb=mb-1
          fwgtb=1.0
        else
          fwgtb=findxb-indxb
        endif
c	Combine everything
	point1=field(indxa,indxb)*(1.0-fwgtb)+
     %     field(indxa,indxb+1)*fwgtb
	point2=field(indxa+1,indxb)*(1.0-fwgtb)+
     %     field(indxa+1,indxb+1)*fwgtb
	getbeta=point1*(1.0-fwgta)+point2*fwgta
c	Derivative divided by ln10*a because of logarithmic scale
	deriv=(point2-point1)/(da*ln10*a)
	return
	end

c	***********************************************************
c	subroutine to initialze the beta field
c	***********************************************************
	subroutine initbeta
	implicit none
	integer ma,mb

	include 'fsizes.inc'
	real field(mx,mx), a0,da,b0,db

	common /betafeld/ a0,da,ma,b0,db,mb,field

11	format(' Escape probability field initialized: ',I3,'x',I3)
	a0=-2.0
	da=0.1
	ma=41
	b0=-2.0
	db=0.08
	mb=51
	call betafield(a0,da,ma,b0,db,mb,field)
	write(6,11) ma,mb
	return
	end

c	***********************************************************
c	subroutine to initailize the field of possible 
c	escape probabilities
c	***********************************************************
	subroutine betafield(a0,da,ma,b0,db,mb,field)
	implicit  none
	integer ma,mb

	include 'fsizes.inc'
	real field(mx,mx),a0,da,b0,db,a,b,integral,limit
	integer i,j,mg

	data mg,limit /1001,4.0/

c	Initialize functions
	call initphi(mg,limit)

	do 100 i=1,ma
	a=10.0**(a0+(i-1)*da)
	do 100 j=1,mb
	b=b0+(j-1)*db
	field(i,j)=integral(a,b)
100	continue
	return
	end

c	***********************************************************
c	Crude integration
c	***********************************************************
	real function integral(a,b)
	implicit none
	real a,b
	integer mgmax

	parameter (mgmax=1001)
	real gm(mgmax),phi(mgmax),xbas,xdif
	real x0,tau0,wgt
	real xbar,xindex,xrel,gmbar,phibar
	real*8 count,all
	integer num,i,ind

	common /gamma/ num,xbas,xdif,gm,phi
	save /gamma/

	wgt=xdif
	all=0d0
c	**********************************************************
c	Different treatment for B<>0 (only for beta)
c	**********************************************************
	if (b.ge.(-1.0/num)) then
c	Outer loop
	do 100 i=1,num
	x0=xbas+(xdif*i)
	tau0=a*(1.0-gm(i)-b*(x0-x0*gm(i)-0.5*phi(i)))
	count=phi(i)*exp(-tau0)
	all=all+count
100	continue
c	**********************************************************
c	Compute x-bar and its functions
c	**********************************************************
	else
	do 200 i=1,num
	x0=xbas+(xdif*i)
	xbar=x0-1.0/b
	xindex=(xbar-xbas)/xdif
	ind=xindex
	if (ind.ge.num) then
	gmbar=1.0
	phibar=0.0
	else
	xrel=xindex-ind
	gmbar=gm(ind)*(1.0-xrel)+gm(ind+1)*xrel
	phibar=phi(ind)*(1.0-xrel)+phi(ind+1)*xrel
	endif
c	New tau's
	tau0=a*(gmbar-gm(i)-b*(x0*gmbar-x0*gm(i)+0.5*(phibar-phi(i))))
	count=phi(i)*exp(-tau0)
	all=all+count
200	continue
	endif
c	**********************************************************
	integral=all*wgt
	return
	end

c	***********************************************************
c	Initialize phi
c	***********************************************************
	subroutine initphi(mg,limit)
	integer j,mg,mgmax,num

	parameter (mgmax=1001)
	real gm(mgmax),phi(mgmax),limit,xbas,xdif,x,sqpi

	common /gamma/ num,xbas,xdif,gm,phi
	save /gamma/
	data sqpi/ 0.5641896/

	num=min(mg,mgmax)
	xdif=2.0*abs(limit)/(num-1)
	xbas=-abs(limit)-xdif

	phi(1)=sqpi*exp(-(xbas+xdif)**2)
	gm(1)=phi(1)*xdif*0.5
	do 100 j=2,num
	x=xbas+(xdif*j)
	phi(j)=sqpi*exp(-x**2)
	gm(j)=gm(j-1)+(phi(j)+phi(j-1))*xdif*0.5
100	continue
c	write(6,*) 'Test integral: ',gm(num/2+1),gm(mg)
	return
	end

