c	**********************************************************
c	Subroutine to estimate the level populations from the
c	velocity profile in Sobolev approximation
c	An appropriate first guess has to be given !
c	**********************************************************
	subroutine sobolev(dhdens,t,vgrad,bgrad,
     %                                   uext,lin,dd,lev,imx)
	implicit none
	integer lin,lev,imx,direct

	include 'fsizes.inc'
	parameter(direct=6)
	real*8 dlocal(nlev), dd(nlev)
	real dhdens,t

	real vgrad(direct),bgrad(direct),dvdr(direct)
	real uext(ntra),uu(ntra),abu(ntra),abu1(ntra)
	real cap,dcap,dcap1,s,ds,ds1,beta,abbeta,rerr
	real conv,neglev
	integer iter,l,k,levtmp,redlev

	common /acclev/ conv,neglev
	save /acclev/

c	**********************************************************
c	Treat all density points from the table
c	**********************************************************
c	Translate from gradients to absolute values
	do 110 k=1,direct
	dvdr(k)=abs(vgrad(k))
c	dvdr(k)=vgrad(k)
110	continue
	do 300 l=1,lev
	dlocal(l)=dd(l)
300	continue

c	**********************************************************
c	Iteration to find the local solution
c	**********************************************************
	iter=0
	redlev=0
	levtmp=lev
400	continue

c	**********************************************************
c	Loop for the levels
c	Compute kappa and S, reduce kappa for turbulent clumping,
c	integrate finally the energy densities
c	**********************************************************
	do 450 l=1,lin
	call sobkapsrc(dlocal,cap,dcap,dcap1,s,ds,ds1,l)

	call gaussob(dvdr,bgrad,cap,beta,abbeta)
	uu(l)  =(1.0-beta)*s + beta*uext(l)
	abu(l) =(1.0-beta)*ds + (abbeta*dcap)*(uext(l)-s)
	abu1(l)=(1.0-beta)*ds1 + (abbeta*dcap1)*(uext(l)-s)
450	continue
c	**********************************************************
c	Set up the matrix and solve the system 
c	Look for convergence
c	**********************************************************
	call sobmatrix(t,dhdens,uu,abu,abu1,dlocal,lin,levtmp,rerr)
	iter=iter+1
	if (iter.ge.imx) goto 500
c	************************************************************
c	Reduce the level number temporarily if necessary
c	Neglect level if unpopulated three subsequent times
c	************************************************************
	if (dlocal(levtmp).le.0.01*neglev) then
	redlev=redlev+1
	if (redlev.ge.3) then
	levtmp=levtmp-1
	redlev=0
	endif
	else
	redlev=0
	endif
	if (rerr.ge.conv) goto 400
c	************************************************************
c	End of the iteration, assign the local values
c	to the global field of level populations
c	************************************************************
500	imx=iter
	do 550 l=1,lev
550	dd(l)=dlocal(l)
	return
	end

c	*************************************************************
c	Subroutine to carry out the mu-integration for the
c	Sobolev approximation, beta and dbeta/dkappa are treated
c	The points and weightings from gauleg are used.
c	The symmetry of the escape probability at 0 is expoited.
c	*************************************************************
	subroutine gaussob(dvdr,bgrad,kappa,y1,y2)
	implicit none
	integer direct

	parameter(direct=6)
	real dvdr(direct),bgrad(direct),kappa,y1,y2,x1,x2
	integer i

	y1=0.0
	y2=0.0 
	do  500  i = 1, direct
	call newescape(dvdr(i),bgrad(i),kappa, x1, x2)
	y1=y1+x1
	y2=y2+x2
500	continue
c	correct weighting
	y1=y1/direct
	y2=y2/direct
	return
	end

c	***************************************************************
c	Subroutine to get the escape probability and its derivative
c	with respect to kappa from the precomputed table
c	***************************************************************
	subroutine newescape(dvdr,bgrad,kappa,beta,abbeta)
	implicit none
	real dvdr,bgrad,kappa,beta,abbeta
	real tau,ableitung,getbeta

c	tau=kappa*dvdr
	tau=kappa/dvdr
c	***************************************************************
c	avoid the accuracy problems for small tau
c	***************************************************************
	if (tau.lt.0.01) then
	beta = 1.0 - tau*0.5
c	abbeta = (tau*0.5-0.5)*dvdr
	abbeta = (tau*0.5-0.5)/dvdr
	else
	beta = getbeta(tau,bgrad,ableitung)
c	abbeta = ableitung*dvdr
	abbeta = ableitung/dvdr
	endif
c	***************************************************************
	return
	end

c	***************************************************************
c	Subroutine to compute the escape probability and its 
c	derivative from a given point into a certain direction
c	beta = (1-exp(-kappa/dvdr)) / (kappa/dvdr)
c	dvdr has a sign, q has none
c	abbeta = dbeta/dkappa
c	***************************************************************
	subroutine escape(dvdr,kappa,beta,abbeta)
	implicit none
	real dvdr,kappa,beta,abbeta
	real tau,exptau

c	***************************************************************
c	Check for zero gradients
c	***************************************************************
	if (dvdr.eq.0.0) then
	beta=0.0
	abbeta=0.0
	return
	endif	

c	tau=kappa*dvdr
	tau=kappa/dvdr
c	***************************************************************
c	avoid the accuracy problems for small tau
c	***************************************************************
	if (tau.lt.0.01) then
	beta = 1.0 - tau*0.5
c	abbeta = (tau*0.5-0.5)*dvdr
	abbeta = (tau*0.5-0.5)/dvdr
	else
	exptau = exp(-tau)
	beta = (1.0-exptau) / tau
c	abbeta = (exptau*(1.0+tau)-1.0)*dvdr / (tau**2)
	abbeta = (exptau*(1.0+tau)-1.0) / (dvdr*tau**2)
	endif
c	***************************************************************
	return
	end

c	*************************************************************
c	Subroutine to compute the derivatives of the absorption 
c	and the source function from the level populations
c	at one radial point
c	*************************************************************
	subroutine sobkapsrc(dd,cap,dcap,dcap1,s,ds,ds1,l)
	implicit none
	integer l

	include 'fsizes.inc'
	real*8 dd(nlev)
	real cap,dcap,dcap1,s,ds,ds1
	real bij,bji,aij,pc
	real hep,lgth,rat,mrat
	integer lfrom(ntra),lto(ntra),l1,l2

        common /lintable/ lfrom, lto
	save /lintable/

c	The length translation to pc is done via kappa
	data pc /3.0856e18/

c	Constants for transition
	lgth=pc
	l1=lfrom(l)
	l2=lto(l)
c	Finite tau
	mrat=1d-15*bji(l)
	rat=dd(l2)*bji(l)-dd(l1)*bij(l)
	if (abs(rat).lt.mrat) rat=mrat
c 	Values and gradients
	cap=lgth*rat
	dcap=lgth*bji(l)
	dcap1=-lgth*bij(l)
	hep=aij(l)/rat
	s=hep*dd(l1)
	ds=-s*(dcap/cap)
	ds1=hep - s*(dcap1/cap)
	return
	end

c	*************************************************************
c	Computation of the new level populations from the local
c	radiation field in the Sobolev approximation
c	The level populations are normalized to 1 here!
c	*************************************************************
c	At first the matrix of transitional balance equations
c	is set up. Then it is translated to differences of level
c	populations. Finally it is solved by lusolver
c	************************************************************* 
	subroutine sobmatrix(T,sdens,uu,abu,abu1,vek,lin,lev,rerr)
	implicit none
	integer lev,lin
	real T,sdens,rerr

	include 'fsizes.inc'
	real uu(ntra),abu(ntra),abu1(ntra)
	real*8 vek(nlev),v(nlev)
	real*8 am(nlev,nlev),sumcol(nlev),urel,dens,rate
	real aij,bij,bji,cij
	integer j,k,kup,kdo
	integer lfrom(ntra),lto(ntra)

        common /lintable/ lfrom, lto
        save /lintable/

	dens=sdens
c	***********************************************************
c	Compute the original matrix
c	Same approach like in matixup
c	***********************************************************
c	Initialize the matrix with the constructive collisional 
c	transitions ( col(j,j)=0 has to be fulfilled )
c	*************************************************************
	do j=1,lev
	sumcol(j)=0d0
	do k=1,lev
	  am(k,j)=dens*cij(T,j,k)
	  sumcol(j)=sumcol(j)+am(k,j)
	enddo
	enddo
	do j=1,lev
	  am(j,j)=am(j,j)-sumcol(j)
	enddo
c	*************************************************************
c	Set up the different lines of the matrix  with radiative
c	transitions. Here, I go through the transitions, not levels.
c	*************************************************************
	do j=1,lin
	  kup=lfrom(j)
	  if (kup.le.lev) then
	  kdo=lto(j)
	  rate=aij(j)+bij(j)*uu(j)
	  am(kdo,kup)=am(kdo,kup)+rate
	  am(kup,kup)=am(kup,kup)-rate
	  rate=bji(j)*uu(j)
	  am(kup,kdo)=am(kup,kdo)+rate
	  am(kdo,kdo)=am(kdo,kdo)-rate
	  endif
	enddo
c	***********************************************************
c	The right hand side
c	***********************************************************
	do j=1,lev-1
	  v(j)=0d0
	  do k=1,lev
	    v(j)=v(j)-vek(k)*am(j,k)
	  enddo
	enddo
c	Last line
	v(lev)=1d0
	do k=1,lev 
	  v(lev)=v(lev)-vek(k)
	enddo
c	***********************************************************
c	Add the derivatives to the matrix
c	***********************************************************
	do j=1,lin
	  kup=lfrom(j)
	  if (kup.le.lev) then
	  kdo=lto(j)
	  rate=abu1(j)*(bij(j)*vek(kup)-bji(j)*vek(kdo))
	  am(kdo,kup)=am(kdo,kup)+rate
	  am(kup,kup)=am(kup,kup)-rate
	  rate=abu(j)*(bji(j)*vek(kdo)-bij(j)*vek(kup))
	  am(kup,kdo)=am(kup,kdo)+rate
	  am(kdo,kdo)=am(kdo,kdo)-rate
	  endif
	enddo
c	*************************************************************
c	The last line. It is not linear independent from the 
c	others when taking the transition equations.
c	Here, the normalization condition has to be used.
c	*************************************************************
	do j = 1,lev
	  am(lev,j) = 1d0 
	enddo
c	***********************************************************
c	The solver
c	***********************************************************
	call lusolver(am,lev,v)
c	***********************************************************
c	Avoid negative populations which may occur during the
c	Newton-Raphson fit , compute the maximum change
c	***********************************************************
	do j=1,lev
	  vek(j)=vek(j)+v(j)
	  if (vek(j).lt.1d-14) vek(j)=1d-14
	enddo
	rerr=0.0d0
	do j=1,lev-2
	  urel=abs(v(j))/vek(j)
	  if (urel.gt.rerr) rerr=urel
	enddo
	return
	end

