c	***********************************************************
c	Subroutine to call the radiative transfer for different
c	total depths up to the considered point and average
c	***********************************************************
	subroutine stplev(dco,vz,dlb,inx,cptb,sctb,uout,uext,
     %			deltar,dny,dztot,nx,ny,nz,lev,nff)
	implicit none
	integer nff,nx,ny,nz,lev

        include 'fsizes.inc'
	real vz(ns,ns,ns),dlb(ns,ns,ns),dco(ns,ns,ns)
	real cptb(ntab,ntra),sctb(ntab,ntra)
	real deltar,uext(ntra),dny,dztot
	real uout(-nf:nf,ntra)
	integer inx(ns,ns,ns)

	real iout(-nf:nf,nlev-1),dz,z0
	integer f,l

c	***********************************************************
c	Call the transfer twice: 1/4 of total way and 3/4
c	***********************************************************
	dz=0.25*dztot
	z0=0.0
	call stpmic(dco,vz,dlb,inx,cptb,sctb,iout,uext,
     %			deltar,dny,z0,dz,nx,ny,nz,lev,nff)
	do 100 f=-nff,nff
	do 100 l=1,lev
100	uout(f,l)=iout(f,l)
	dz=0.75*dztot
	z0=0.25*dztot
	call stpmic(dco,vz,dlb,inx,cptb,sctb,iout,uext,
     %			deltar,dny,z0,dz,nx,ny,nz,lev,nff)
	do 200 f=-nff,nff
	do 200 l=1,lev
200	uout(f,l)=0.5*uout(f,l)+0.5*iout(f,l)
	return
	end

c	***********************************************************
c	Subroutine to carry out radiative transport computation
c	It determines the energy density at all given radial points
c	***********************************************************
	subroutine stpmic(dco,vz,dlb,inx,cptb,sctb,iout,uext,
     %			deltar,dny,z0,dztot,nx,ny,nz,lev,nff)
	implicit none
	integer nff,nx,ny,nz,lev

	include 'fsizes.inc'
	real vz(ns,ns,ns),dlb(ns,ns,ns),dco(ns,ns,ns)
	real cptb(ntab,ntra),sctb(ntab,ntra)
	real deltar,uext(ntra),dny,dztot,z0
	real iout(-nf:nf,ntra)
	integer inx(ns,ns,ns)

	real cap(-nf:nf,ntra),iplus(-nf:nf,ntra)
	real cpo(ntra),sco(ntra),cp(ntra),sca(ntra)

	real vs,dv,vso,dvo,z
	integer i,j,l,j1,j2,it,nztot,nz0,f
	integer kk,itstp,kstp

c	********************************************************
c	Get the number of total steps through the cloud
c	********************************************************
	nz0=int(z0/deltar)+1 
	nztot=int(dztot/deltar)+nz0 
	do 110 l=1,lev
	do 110 j=-nff,nff
110	iout(j,l)=0.0

c	********************************************************
c	This is the beginning of the loop over the whole field 
c	********************************************************
	do 400 i=1,nx
	do 410 j=1,ny

c	********************************************************
c	Beginning of the z-loop (integration along a ray)
c	********************************************************
c	Special treatment for the most outer left point
c	********************************************************
c	Geometry parameters
	kk=nz0
	vso = vz(i,j,kk)
	dvo = dlb(i,j,kk)
c	Background for all frequencies
	do 500 l=1,lev
	do 500 f=-nff,nff
500	iplus(f,l) = uext(l)
c	Initialize kappa and source function fields
	do 510 l=1,lev
	cpo(l)=dco(i,j,kk)*cptb(inx(i,j,kk),l)
510	sco(l)=sctb(inx(i,j,kk),l)
	call leftkpm(cpo,cap,j1,j2,lev,vso,dvo,dny)

c	********************************************************
c	Now integration - treat all points through the cloud
c	********************************************************
	do  1600  kstp = nz0+1,nztot
	kk=mod(kstp-1,nz)+1
c	Initialize geometry parameters
	vs= vz(i,j,kk)
	dv = dlb(i,j,kk)
	do 610 l=1,lev
	cp(l)=dco(i,j,kk)*cptb(inx(i,j,kk),l)
610	sca(l)=sctb(inx(i,j,kk),l)
c	*******************************************************
c	Radiative transfer - Multiple steps if the velocity
c	gradient is too large
c	*******************************************************
	call getstpm(vso,dvo,cpo,sco,vs,dv,cp,sca,deltar,z,lev,itstp)
c	Jumps
	if (itstp.eq.0) then
	call transfer(cpo,sca,iplus,cap,sco,j1,j2,lev,vso,dvo,dny,z)
	call nextstpm(vso,dvo,cpo,sca,lev)
	do 620 l=1,lev
620	sco(l)=sca(l)
	call leftkpm(cpo,cap,j1,j2,lev,vso,dvo,dny)
	call transfer(cpo,sca,iplus,cap,sco,j1,j2,lev,vso,dvo,dny,z)
	else
	do 640 it=1,itstp
	call nextstpm(vso,dvo,cpo,sca,lev)
	call transfer(cpo,sca,iplus,cap,sco,j1,j2,lev,vso,dvo,dny,z)
640	continue
	endif
1600	continue
c	**********************************************************
c	End of the z loop
c	Store the results into a large field
c	**********************************************************
	do 420 l=1,lev
	do 420 f=-nff,nff
	iout(f,l) = iout(f,l)+iplus(f,l)
420	continue
410	continue
400	continue
c	**********************************************************
c	Average from sum
c	**********************************************************
	do 440 l=1,lev
	do 440 f=-nff,nff
440	iout(f,l)=iout(f,l)/(nx*ny)
	return
	end

c	************************************************************
c	Subroutine to compute the physical parameters at the end of 
c	the first radiative transfer step within one interval.
c	It checks how many steps have to be done there.
c	This is a reduced version of firststep
c	************************************************************
	subroutine getstpm(vso,dvo,cpo,sco,vs,dv,cp,sc,dr,z,lev,istp)
	implicit none
	integer istp,lev

	include 'fsizes.inc'
	real cpo(ntra),sco(ntra),cp(ntra),sc(ntra)
	real vs,vso,dv,dvo,dr,z
	real deltav,vjump,explin,softzone
	real ddcp(ntra),ddsc(ntra),ddv,ddt
	integer l

c	Internal variables are transmitted to nextlstp for all
c	later steps
	common /stepm/ ddcp,ddsc,ddv,ddt
	common /accstep/ deltav,vjump,explin,softzone
	save /stepm/,/accstep/

c	*******************************************************
c	Jump or linear relation
c	*******************************************************
	ddv=vs-vso
	if (abs(ddv).gt.vjump*dv) then
	istp=0
	ddt=(dv-dvo)
	do 610 l=1,lev
	ddcp(l)=(cp(l)-cpo(l))
610	ddsc(l)=(sc(l)-sco(l))
	z=0.25*dr
c	*******************************************************
c	linear approx 
c	*******************************************************
	else
	istp=int(abs(ddv)/(deltav*dv))+1
	ddv=ddv/istp
	ddt=(dv-dvo)/istp
	do 620 l=1,lev
	ddcp(l)=(cp(l)-cpo(l))/istp
620	ddsc(l)=(sc(l)-sco(l))/istp
	z=0.5*dr/istp
	endif
	do 630 l=1,lev
630	sc(l)=sco(l)
	return
	end

c	*********************************************************
c	Subroutine to compute the physical parameters at the
c	end of the following radiative transfer steps
c	*********************************************************
	subroutine nextstpm(vso,dvo,cpo,sco,lev)
	implicit none
	integer lev

	include 'fsizes.inc'
	real cpo(ntra),sco(ntra)
	real vso,dvo
	integer l

c	Internal variables computed in firststep 
	real ddcp(nlev-1),ddsc(nlev-1),ddv,ddt
	common /stepm/ ddcp,ddsc,ddv,ddt
	save /stepm/

c	Update the physical parameters with firststep variables
	do 110 l=1,lev
	cpo(l)=cpo(l)+ddcp(l)
110	sco(l)=sco(l)+ddsc(l)
	vso=vso+ddv
	dvo=dvo+ddt
	return
	end

c	***********************************************************
c	Initialization of the kappa field on the left cloud edge
c	This is a reduced version of leftkp for only one line
c	***********************************************************
	subroutine leftkpm(cp,cappa,j1,j2,lev,vs,dv,dny)
	implicit none
	integer j1,j2,lev

	include 'fsizes.inc'
	real cp(ntra), cappa(-nf:nf,ntra)
	real vs,dv,dny
	real negexp,negint
	real phi,prof
	integer j,l

	common /accgauss/ negexp,negint
	save /accgauss/

c	Frequency range
	j1=int((vs-dv*negexp)/dny)-1
	j2=int((vs+dv*negexp)/dny)+1
c	Frequency dependent quantities
	do 500 j=j1,j2
	prof = phi(j*dny,vs,dv)
	do 500 l=1,lev
	cappa(j,l)  = cp(l) * prof
500	continue
	return
	end

c	***********************************************************
c	Subroutine for radiative transfer between two points
c	***********************************************************
	subroutine transfer(cp,sc,iplus,cap,ss,j1,j2,lev,vs,dv,dny,z)
	implicit none
	integer lev,j1,j2

	include 'fsizes.inc'
	real cp(ntra),sc(ntra),ss(ntra)
	real cap(-nf:nf,ntra)
	real iplus(-nf:nf,ntra), pp(-nf:nf)
	real vs,dv,dny,z
	real negexp,negint,explin,deltav,vjump,softzone
	real phi,prof,tau1,tau2,tau,tauq
	real g1,ex1,ex2
	real ihlp1,ihlp2,rel
	integer jn1,jn2,jj1,jj2,j,l

	common /accstep/ deltav,vjump,explin,softzone
	common /accgauss/ negexp,negint
	save /accstep/,/accgauss/

c	*******************************************************
c	Determine indices of the frequency range
c	*******************************************************
	jn1=int((vs-dv*negexp)/dny)-1
	jn2=int((vs+dv*negexp)/dny)+1
c	Find the overlapping interval
	jj1=max(j1,jn1)
	jj2=min(j2,jn2)
c	*******************************************************
c	Treat the non-overlapping left end
c	*******************************************************
	if (jn1.lt.jj1) then
	do 510 j=jn1,jj1-1
	prof = phi(j*dny,vs,dv)
	do 510 l=1,lev
	cap(j,l) = cp(l)*prof
	tau = z*cap(j,l)
	iplus(j,l)=(iplus(j,l)+sc(l)*tau)/(1.0+tau)
510	continue
	else if (j1.lt.jj1) then
	do 520 j=j1,jj1-1
	do 520 l=1,lev
	tau = z*cap(j,l)
	iplus(j,l)=(iplus(j,l)+ss(l)*tau)/(1.0+tau)
520	continue
	endif
c	******************************************************
c	Treat the main, fully overlapping part
c	Only in this part, it is controlled whether the lines 
c	are optically thick, i.e. it is assumed that they are
c	thin within the non-overlapping wings
c	******************************************************
	prof=2.0*z*phi(0.0,0.0,dv)
	do 500 j=jj1,jj2
500	pp(j) = phi(j*dny,vs,dv)
	do 600 l=1,lev
c	******************************************************
c	When even the centre of the line is optically 
c	thin the time consuming if in the inner loop shall
c	be avoided
c	******************************************************
c	Starting from SimLine 2.0, I prevent exponential maser
c	amplification for stability reasons. This is unphysical
c	in situations with real masers - no final solution yet.
c	Uncomment the old lines below for maser amplification.
c	******************************************************
c	if (abs(cp(l)*prof).lt.explin) then
	if (cp(l)*prof.lt.explin) then
	do 650 j=jj1,jj2
	tau1 = z*cap(j,l)
	cap(j,l) = cp(l)*pp(j)
	tau2 = z*cap(j,l)
	tauq = tau1*tau2
	iplus(j,l) = (iplus(j,l)+sc(l)*tau2+ss(l)*tau1
     %	 +0.66667*tauq*(2.0*sc(l)+ss(l)))/(1.0+tau1+tau2+2.0*tauq)
650	continue
	else
c	******************************************************
c	In each step there is independently asked whether
c	the line is optically thick
c	It has to be guaranteed that none of the cappa values
c	is 0 in this case. (Should be always fulfilled for
c	the overlapping part.)
c	******************************************************
	do 680 j=jj1,jj2
	tau1 = z*cap(j,l)
	cap(j,l) = cp(l)*pp(j)
	tau2 = z*cap(j,l) 
	tau = tau1+tau2
c	rel = abs(tau)
c	if (rel.lt.explin) then
	if (tau.lt.explin) then
c	Thin case - linear approx.
	tauq = tau1*tau2
	iplus(j,l) = (iplus(j,l)+sc(l)*tau2+ss(l)*tau1
     %	 +0.66667*tauq*(2.0*sc(l)+ss(l)))/(1.0+tau+2.0*tauq)
c	else if (rel.lt.softzone) then
	else if (tau.lt.softzone) then
c	Small transition regime - superposition
	tauq=tau1*tau2
	ihlp1 = (iplus(j,l)+sc(l)*tau2+ss(l)*tau1
     %	 +0.66667*tauq*(2.0*sc(l)+ss(l)))/(1.0+tau+2.0*tauq)
	ex1=exp(-0.33333*tau1)
	ex2=exp(-0.33333*tau2)
	g1=0.5*(ss(l)-sc(l))
	ihlp2=sc(l)+ex2*ex2*(g1+ex2*ex1*
     %              (g1+ex1*ex1*(iplus(j,l)-ss(l))))
c	iplus(j,l)=(ihlp1*(softzone-rel)+ihlp2*(rel-explin))/
c     %              (softzone-explin)
	iplus(j,l)=(ihlp1*(softzone-tau)+ihlp2*(tau-explin))/
     %              (softzone-explin)
	else
c	Optically thick case - exponential regime
	ex1=exp(-0.33333*tau1)
	ex2=exp(-0.33333*tau2)
	g1=0.5*(ss(l)-sc(l))
	iplus(j,l)=sc(l)+ex2*ex2*(g1+ex2*ex1*
     %              (g1+ex1*ex1*(iplus(j,l)-ss(l))))
	endif
680	continue
	endif
600	continue
c	*******************************************************
c	Treat the non-overlapping right end
c	*******************************************************
	if (jn2.gt.jj2) then
	do 560 j=jj2+1,jn2
	prof = phi(j*dny,vs,dv)
	do 560 l=1,lev
	cap(j,l) = cp(l)*prof 
	tau = z*cap(j,l)
	iplus(j,l)=(iplus(j,l)+sc(l)*tau)/(1.0+tau)
560	continue
	else if (j2.gt.jj2) then
	do 570 j=jj2+1,j2
	do 570 l=1,lev
	tau = z*cap(j,l)
	iplus(j,l)=(iplus(j,l)+ss(l)*tau)/(1.0+tau)
570	continue
	endif
c	*******************************************************
c	New parameters
c	*******************************************************
	j1=jn1
	j2=jn2
	do 580 l=1,lev
580	ss(l)=sc(l)
	return
	end

